// Cre par David Croft pour le cours d'IA a l'ESI
// fvrier 2001

package edu.esi.lettresdespace.prof;

import edu.esi.lettresdespace.*;
import edu.esi.outils.Debug;
import java.util.*;

/**
 * This class is used to generate a new video containing a hidden message. 
 * The idea is to load a set of sample letters, and then somewhat randomnly
 * (and yet still maintaining the correct sequence for the message) insert 
 * letters into the series of video frames (Cadres). All video frames 
 * (those that represent letters and those that represent nothing) have
 * random noise added to them.
 */
public class VideoCreateur {

	private static final int PERCENTAGE_BRUYANT_LETTRE = 8;
	private static final int PERCENTAGE_BRUYANT_VIERGE = 50;
	
	/**
	 * Used to keep track of the samples for each letter. For example, 
	 * for the letter A, we might have 2 or more samples that we want 
	 * to randomnly choose between each time we insert a letter A into 
	 * the video.
	 */
	private class Lettre {
		char lettre_;
		Vector enchantillonsDeLaLettre_ = new Vector(5);

		Lettre(char letter) {
			lettre_ = letter;
		}
		public void addEnchantillon(Cadre cadre) {
			Debug.ASSERT(cadre.getLettre() == lettre_,
				"Lettre dans le cadre ce n'est pas la meme que cette groupe " +
				"des enchantillons");
			enchantillonsDeLaLettre_.addElement(cadre);
		}
		Cadre getRandom() {
			int numEnchantillons = enchantillonsDeLaLettre_.size();
			int randomIndex = random_.nextInt(numEnchantillons-1);
			return (Cadre)enchantillonsDeLaLettre_.elementAt(randomIndex);
		}
	}

	private Hashtable lettreTable_ = new Hashtable(250);
	private Random random_ = new Random();
	
	/**
	 * Create a new VideoCreateur - requires the name of the folder and the
	 * file that contains the samples for all of the letters. (We need to 
	 * have atleast one sample for each type of letter (eg. 'A') contained in
	 * the message. If the message contains letters for which we have no
	 * sample, the video cannot be created.
	 */
	public VideoCreateur(	String dossier, 
							String nomDeLaFichier) {
		
		Vector lettres = CadreChargeur.charge(dossier, nomDeLaFichier);
		Enumeration e = lettres.elements();
		while (e.hasMoreElements()) {
			Cadre nextCadre = (Cadre)e.nextElement();
			Character cadreChar = new Character(nextCadre.getLettre());
			Lettre lettre = null;
			if (lettreTable_.containsKey(cadreChar)) {
				lettre = (Lettre)lettreTable_.get(cadreChar);
			} else {
				lettre = new Lettre(cadreChar.charValue());
				lettreTable_.put(cadreChar, lettre);
			}
			lettre.addEnchantillon(nextCadre);
		}
	}
	
	/**
	 * Create a new video with the specified number of frames and 
	 * randomnly insert the message into this series of frames.
	 */
	public Video cree(String message, int numCadres) {
		
		// current we only support upper case letters 
		// (seulement majiscules!)
		message = message.toUpperCase();

		// replace all spaces with underscores
		message = message.replace(' ','_');

		System.out.println("Cree une video avec " + numCadres +
			" cadres pour le message:\n" + message);
		
		// first, we want to figure out the sequence where we will place
		// the frames representing letters
		// (message.length() characters between 0 and numCadres)
		
		int numLettres = message.length();
		System.out.println(numLettres + " caracters dans la message");

		Debug.ASSERT(numCadres >= numLettres,
			"Il faut qu'il y a plus des cadres que lettres dans le message");
		
		random_ = new Random();
		Cadre[] cadres = new Cadre[numCadres];
		int interval = numCadres/numLettres;
		int currentInterval = 0;

		for (int i=0; i < numLettres; i++) {
			int nextLettreIndex = currentInterval + random_.nextInt(interval);
			for (int j=currentInterval; j < currentInterval+interval; j++) {
				if (j == nextLettreIndex) {
					Character next = new Character(message.charAt(i));

					Lettre lettre = (Lettre)lettreTable_.get(next);
					Debug.ASSERT(lettre,
						"Il n'y a pas une lettre registe pour caracter: '" +
						next.toString() + "'");

					Cadre cadre = lettre.getRandom();
					Cadre noisyCadre = new Cadre(cadre);
					addNoise(noisyCadre, PERCENTAGE_BRUYANT_LETTRE);
					cadres[j] = noisyCadre;
				} else {
					cadres[j] = newCadreViergeBruyant();
				}
			}
			currentInterval += interval;
		}
		
		// check that we haven't left any null frames at the end 
		// and remove any letter labels from the frames
		for (int i=0; i < numCadres; i++) {
			if (cadres[i] == null) {
				cadres[i] = newCadreViergeBruyant();
			}
			// make sure we remove the letter marks from all the cadres 
			// - otherwise there is no need to use a neural network
			// to recognize what character a cadre might represent!
			cadres[i].setLettre(Character.MIN_VALUE);
		}
		return new Video(cadres);
	}

	private final Cadre newCadreViergeBruyant() {
		Cadre cadre = new Cadre();
		addNoise(cadre, random_.nextInt(PERCENTAGE_BRUYANT_VIERGE)); 
		return cadre;
	}

	private final void addNoise(Cadre cadre, int percentage) {
		
		// eg. 64
		int length = Cadre.HAUTEUR * Cadre.LARGEUR;
			
		// eg. 64 * percentage
		int maxNumberElementsBruyant = (int) 
					(((float)length) *
					(((float)percentage)/100.0f));

		// Random.nextInt() requires a number > 0
		if (maxNumberElementsBruyant == 0) {
			maxNumberElementsBruyant = 1;
		}

		int numElementsBruyant = random_.nextInt(maxNumberElementsBruyant);

		// Random.nextInt() requires a number > 0
		if (numElementsBruyant == 0) {
			numElementsBruyant = 1;
		}

		// eg. 64/5 = 12
		int interval = length/numElementsBruyant;
		int currentInterval = 0;	

		for (int i=0; i < numElementsBruyant; i++) {
			// choose a new random spot 
			int position = currentInterval + random_.nextInt(interval);
			double newValue = 1.0;
			double currentValue = cadre.getDoubleAt(position);
			if (currentValue == 1.0) {
				newValue = 0.0;
			}
			cadre.set(position, newValue);
			currentInterval += interval;
		}
	}
}
